from datetime import datetime

import pytest

from chalk.features import feature, feature_time, features
from chalk.features.feature import unwrap_feature


@features(tags="t", owner="andy@chalk.ai")
class WowFS:
    # This is a really neat description of something
    something: str

    # This is also really neat and cool
    something_else: str
    nocomment: str

    nope: str

    assigned: str = feature(tags="a")

    # bizarre
    bizarre: str

    # goofy
    goofy: str = feature(owner="yo")

    # now with feature
    assigned_comment: str = feature(tags=["3", "4"])

    # implicit comment
    explicit: str = feature(description="explicit comment")

    # Multiline
    # Neat and, verily, cool
    #
    # Hello
    assigned_comment_multiline: str = feature()

    # now with feature time
    time: datetime = feature_time()

    nope_nope: datetime  # Datetime field that is not feature time


@features(owner="elliot@chalk.ai")
class OwnerFeatures:
    plain: str
    cached: str = feature(max_staleness="3d")
    andy: str = feature(owner="andy@chalk.ai")

    ft: datetime = feature_time()


@features(tags=["1", "2"])
class TagFeatures:
    empty: str
    one: str = feature(tags="one")
    many: str = feature(tags=["a", "b"])

    ft: datetime = feature_time()


@features
class CommentBaseOwner:
    # I'm a cool comment!
    # :owner: elliot@chalk.ai
    empty: str

    # I'm a cool comment!
    # :tags: pii group:risk
    email: str

    # :tags: pii, group:risk
    email_commas: str
    # :tags: pii
    email_single: str

    # :tags: pii
    email_all_kinds: str = feature(tags=["hello"])


def test_comment_based_owner():
    assert "elliot@chalk.ai" == unwrap_feature(CommentBaseOwner.empty).owner
    assert ["pii", "group:risk"] == unwrap_feature(CommentBaseOwner.email).tags
    assert ["pii"] == unwrap_feature(CommentBaseOwner.email_single).tags
    assert ["hello", "pii"] == unwrap_feature(CommentBaseOwner.email_all_kinds).tags
    with pytest.raises(ValueError):

        @features
        class BadFeatureClass:
            # :owner: elliot@chalk.ai
            doubly_owned: str = feature(owner="d")


def test_parse_descriptions():
    assert "bizarre" == unwrap_feature(WowFS.bizarre).description
    assert "This is a really neat description of something" == unwrap_feature(WowFS.something).description
    assert "explicit comment" == unwrap_feature(WowFS.explicit).description
    assert "This is also really neat and cool" == unwrap_feature(WowFS.something_else).description
    assert "now with feature" == unwrap_feature(WowFS.assigned_comment).description
    assert "goofy" == unwrap_feature(WowFS.goofy).description
    assert (
        """Multiline
Neat and, verily, cool

Hello"""
        == unwrap_feature(WowFS.assigned_comment_multiline).description
    )


def test_class_owner():
    assert "elliot@chalk.ai" == OwnerFeatures.__chalk_owner__
    assert "elliot@chalk.ai" == unwrap_feature(OwnerFeatures.plain).owner
    assert "elliot@chalk.ai" == unwrap_feature(OwnerFeatures.cached).owner
    assert "andy@chalk.ai" == unwrap_feature(OwnerFeatures.andy).owner


def test_class_tags():
    assert ["1", "2"] == TagFeatures.__chalk_tags__
    assert ["1", "2"] == unwrap_feature(TagFeatures.empty).tags
    assert ["one", "1", "2"] == unwrap_feature(TagFeatures.one).tags
    assert ["a", "b", "1", "2"] == unwrap_feature(TagFeatures.many).tags
    assert ["a", "t"] == unwrap_feature(WowFS.assigned).tags
    assert ["t"] == unwrap_feature(WowFS.nope).tags
    assert ["3", "4", "t"] == unwrap_feature(WowFS.assigned_comment).tags



import enum
from datetime import date, datetime
from typing import List, Set

import pendulum

from chalk.features import feature, features
from chalk.features.feature import unwrap_feature
from chalk.serialization.codec import FeatureCodec


class CustomClass:
    def __init__(self, w: str):
        self.w = w

    def __eq__(self, other):
        return isinstance(other, CustomClass) and self.w == other.w


class Color(enum.Enum):
    blue = "blue"
    green = "green"
    orange = "orange"


@features
class Hello:
    a: str
    b: int
    c: datetime
    d: Color
    e: date
    y: Set[int]
    z: List[str]
    fancy: CustomClass = feature(encoder=lambda x: x.w, decoder=lambda x: CustomClass(x))


def _check_roundtrip(f, value):
    codec = FeatureCodec()
    f = unwrap_feature(f)
    encoded = codec.encode(f, value)
    decoded = codec.decode(f, encoded)
    assert decoded == value


def test_datetime():
    codec = FeatureCodec()
    serialized = "2022-04-08T22:26:03.303000+00:00"
    decoded = codec.decode(unwrap_feature(Hello.c), serialized)
    assert decoded == pendulum.parse(serialized)
    re_encoded = codec.encode(unwrap_feature(Hello.c), decoded)
    assert serialized == re_encoded


def test_custom_codecs():
    _check_roundtrip(Hello.fancy, CustomClass("hihi"))


def test_color():
    _check_roundtrip(Hello.d, Color.green)


def test_date():
    codec = FeatureCodec()
    assert "2022-04-08" == codec.encode(unwrap_feature(Hello.e), date.fromisoformat("2022-04-08"))
    _check_roundtrip(Hello.e, date.fromisoformat("2022-04-08"))


def test_list():
    _check_roundtrip(Hello.z, ["hello", "there"])



import dataclasses
from datetime import datetime
from types import NoneType
from typing import Type, TypeVar

import pydantic

from chalk.features import DataFrame, Feature, Filter
from chalk.features.feature import HasOnePathObj
from chalk.features.resolver import OfflineResolver, OnlineResolver, SinkResolver
from chalk.parsed.duplicate_input_gql import (
    KafkaConsumerConfigGQL,
    UpsertDataFrameGQL,
    UpsertFeatureGQL,
    UpsertFeatureIdGQL,
    UpsertFeatureReferenceGQL,
    UpsertFeatureTimeKindGQL,
    UpsertFilterGQL,
    UpsertHasManyKindGQL,
    UpsertHasOneKindGQL,
    UpsertReferencePathComponentGQL,
    UpsertResolverGQL,
    UpsertResolverOutputGQL,
    UpsertScalarKindGQL,
    UpsertSinkResolverGQL,
    UpsertStreamResolverGQL,
)
from chalk.streams.KafkaSource import KafkaConsumerConfig
from chalk.streams.stream import StreamResolver

T = TypeVar("T")


try:
    import attrs
except ImportError:
    attrs = None


def _get_feature_id(s: Feature):
    assert s.name is not None
    assert s.namespace is not None
    return UpsertFeatureIdGQL(
        fqn=s.fqn,
        name=s.name,
        namespace=s.namespace,
    )


def _convert_df(df: Type[DataFrame]) -> UpsertDataFrameGQL:
    return UpsertDataFrameGQL(
        columns=[_get_feature_id(f) for f in df.columns],
        filters=[convert_type_to_gql(f) for f in df.filters],
    )


def _get_path_component(pc: HasOnePathObj) -> UpsertReferencePathComponentGQL:
    assert isinstance(pc.parent, Feature)
    return UpsertReferencePathComponentGQL(
        parent=_get_feature_id(pc.parent),
        child=_get_feature_id(pc.child),
        parentToChildAttributeName=pc.parent_to_child_attribute_name,
    )


def convert_type_to_gql(t: T):
    if isinstance(t, KafkaConsumerConfig):
        return KafkaConsumerConfigGQL(
            broker=list(t.broker) if isinstance(t.broker, str) else t.broker,
            topic=list(t.topic) if isinstance(t.topic, str) else t.topic,
            sslKeystoreLocation=t.ssl_keystore_location,
            clientIdPrefix=t.client_id_prefix,
            groupIdPrefix=t.group_id_prefix,
            topicMetadataRefreshIntervalMs=t.topic_metadata_refresh_interval_ms,
            securityProtocol=t.security_protocol,
        )

    if isinstance(t, StreamResolver):
        return UpsertStreamResolverGQL(
            fqn=t.fqn,
            kind="stream",
            config=convert_type_to_gql(t.source.consumer_config),
            functionDefinition=t.function_definition,
            environment=[t.environment] if isinstance(t.environment, str) else t.environment,
            doc=t.fn.__doc__,
        )

    if isinstance(t, SinkResolver):
        return UpsertSinkResolverGQL(
            fqn=t.fqn,
            functionDefinition=t.function_definition,
            inputs=[
                UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(f),
                    path=[_get_path_component(p) for p in f.path or []],
                )
                for f in t.inputs
            ],
            environment=t.environment,
            tags=t.tags,
            doc=t.doc,
            machineType=t.machine_type,
            bufferSize=t.buffer_size,
            debounce=t.debounce,
            maxDelay=t.max_delay,
            upsert=t.upsert,
        )

    if isinstance(t, (OnlineResolver, OfflineResolver)):
        assert isinstance(
            t.cron, (NoneType, (NoneType, str))
        ), f"Only supporting cron as a string right now for {t.fqn}"

        return UpsertResolverGQL(
            fqn=t.fqn,
            kind="offline" if isinstance(t, OfflineResolver) else "online",
            functionDefinition=t.function_definition,
            inputs=[
                UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(f),
                    path=[_get_path_component(p) for p in f.path or []],
                )
                for f in t.inputs
            ],
            output=UpsertResolverOutputGQL(
                features=[
                    _get_feature_id(f)
                    for f in t.output.features
                    if not isinstance(f, type) or not issubclass(f, DataFrame)
                ],
                dataframes=[
                    _convert_df(f) for f in t.output.features if isinstance(f, type) and issubclass(f, DataFrame)
                ],
            ),
            environment=t.environment,
            tags=t.tags,
            doc=t.doc,
            cron=t.cron,
            machineType=t.machine_type,
        )

    if isinstance(t, Feature):
        assert t.name is not None
        assert t.namespace is not None
        scalar_kind_gql = None
        has_one_kind_gql = None
        has_many_kind_gql = None
        feature_time_kind_gql = None
        if t.is_has_one:
            has_one_kind_gql = UpsertHasOneKindGQL(join=convert_type_to_gql(t.join))
        elif t.is_has_many:
            has_many_kind_gql = UpsertHasManyKindGQL(
                join=convert_type_to_gql(t.join),
                columns=None,
                filters=None,
            )
        elif t.is_feature_time:
            assert t.typ is not None
            assert issubclass(t.typ.underlying, datetime)
            feature_time_kind_gql = UpsertFeatureTimeKindGQL()
        else:
            assert t.typ is not None
            underlying_type = t.typ.underlying
            assert isinstance(underlying_type, type)
            base_classes = [x.__name__ for x in type.mro(underlying_type)]

            # Attrs and Dataclasses don't technically have base classes
            # Pydantic calls their base class BaseModel which is way too generic for string comparison
            # For simplicity on the server-side validation, we'll come up with our own "base class" names
            if dataclasses.is_dataclass(underlying_type):
                base_classes.append("__dataclass__")
            if attrs is not None and isinstance(underlying_type, type) and attrs.has(underlying_type):
                base_classes.append("__attrs__")
            if issubclass(underlying_type, pydantic.BaseModel):
                base_classes.append("__pydantic__")

            scalar_kind_gql = UpsertScalarKindGQL(
                scalarKind=t.typ.underlying.__name__,
                primary=t.primary,
                baseClasses=base_classes,
                version=t.version,
                hasEncoderAndDecoder=t.encoder is not None and t.decoder is not None,
            )

        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            maxStaleness=t.max_staleness,
            description=t.description,
            owner=t.owner,
            etlOfflineToOnline=t.etl_offline_to_online,
            tags=t.tags,
            hasOneKind=has_one_kind_gql,
            hasManyKind=has_many_kind_gql,
            scalarKind=scalar_kind_gql,
            featureTimeKind=feature_time_kind_gql,
        )

    if isinstance(t, Filter):
        return UpsertFilterGQL(
            lhs=_get_feature_id(t.lhs),
            op=t.operation,
            rhs=_get_feature_id(t.rhs),
        )

    return t



import collections.abc
import dataclasses
import enum
import logging
from datetime import date, datetime, timezone
from typing import Any, Dict, Optional, Type, Union, get_origin

import cattrs
import numpy as np
import pandas
import pendulum
import pydantic
import pytz
from dateutil import parser
from sqlalchemy import Boolean, Float, Integer, Text
from sqlalchemy.types import TypeEngine

from chalk.features import Feature, FeatureSetBase
from chalk.utils.enum import get_enum_value_type
from chalk.utils.sqlalchemy import UtcDateTime

try:
    import attrs
except ImportError:
    # Imports not available. Attrs is not required.
    attrs = None

_log = logging.getLogger(__name__)


class FeatureCodec:
    def __init__(
        self,
        fqn_to_feature: Optional[Dict[str, Feature]] = None,
    ):
        self._fqn_to_feature = (
            fqn_to_feature
            if fqn_to_feature is not None
            else {
                feature.fqn: feature
                for fsb in FeatureSetBase.registry.values()
                for feature in fsb.features
                if isinstance(feature, Feature)
            }
        )
        self.converter = cattrs.Converter()
        self.converter.register_structure_hook(datetime, lambda v, _: parser.isoparse(v))
        self.converter.register_unstructure_hook(
            datetime, lambda v: (v if v.tzinfo else pytz.utc.localize(v)).isoformat()
        )

    def _default_encode(self, value: Any, for_pandas: bool = False):
        if isinstance(value, set):
            return {self._default_encode(x) for x in value}
        if isinstance(value, list):
            return [self._default_encode(x) for x in value]
        if isinstance(value, (str, int, float)):
            return value
        if isinstance(value, enum.Enum):
            return self._default_encode(value.value)
        if isinstance(value, datetime):
            tz = value.tzinfo or (datetime.now(timezone.utc).astimezone().tzinfo)
            return pendulum.instance(value, tz).isoformat()
        if isinstance(value, date):
            return value.isoformat()
        if isinstance(value, np.integer):
            return int(value)
        if isinstance(value, np.floating):
            return float(value)
        if isinstance(value, pandas.Timestamp):
            return pendulum.instance(value.to_pydatetime()).isoformat()
        if isinstance(value, pydantic.BaseModel):
            return value.dict()
        if (attrs is not None and attrs.has(type(value))) or dataclasses.is_dataclass(value):
            return self.converter.unstructure(value)
        if isinstance(value, collections.abc.Iterable):
            return [self._default_encode(x) for x in value]
        raise TypeError(f"Unable to encode value of type {type(value).__name__}")

    def encode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.encoder is not None:
            return feature.encoder(value)

        return self._default_encode(value)

    def encode_fqn(self, fqn: str, value: Any):
        if fqn in self._fqn_to_feature:
            return self.encode(self._fqn_to_feature[fqn], value)

        return self._default_encode(value)

    def _default_decode_value(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if issubclass(feature.typ.underlying, enum.Enum):
            value = feature.typ.underlying(value)
        if issubclass(feature.typ.underlying, datetime):
            value = parser.isoparse(value)
        elif issubclass(feature.typ.underlying, date):
            # note: datetime is a subclass of date, so we must be careful to decode accordingly
            value = date.fromisoformat(value)
        if issubclass(feature.typ.underlying, pydantic.BaseModel):
            return feature.typ.underlying(**value)
        elif (attrs is not None and attrs.has(feature.typ.underlying)) or dataclasses.is_dataclass(
            feature.typ.underlying
        ):
            return self.converter.structure(value, feature.typ.underlying)
        if not isinstance(value, feature.typ.underlying):
            raise TypeError(f"Unable to decode value {value} to type {feature.typ.underlying.__name__}")
        return value

    def _default_decode(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if value is None:
            if feature.typ.is_nullable:
                return None
            else:
                raise ValueError(f"Value is none but feature {feature} is not nullable")
        if feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == set:
            if not isinstance(value, set):
                raise TypeError(f"Feature {feature} is a set but value {value} is not")
            return {self._default_decode_value(feature, x) for x in value}
        elif feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == list:
            if not isinstance(value, list):
                raise TypeError(f"Feature {feature} is a list but value {value} is not")
            return [self._default_decode_value(feature, x) for x in value]
        else:
            return self._default_decode_value(feature, value)

    def decode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.decoder is not None:
            return feature.decoder(value)

        return self._default_decode(feature, value)

    def decode_fqn(
        self,
        fqn: str,
        value: Any,
    ):
        if fqn in self._fqn_to_feature:
            return self.decode(self._fqn_to_feature[fqn], value)
        return value

    def get_pandas_dtype(self, fqn: str) -> str:
        feature = self._fqn_to_feature[fqn]
        if feature.pandas_dtype is not None:
            return feature.pandas_dtype
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return "string"
        if issubclass(underlying, bool):
            return "boolean"
        if issubclass(underlying, int):
            return "Int64"
        if issubclass(underlying, float):
            return "Float64"
        if issubclass(underlying, datetime):
            # This assumes timezone-aware. For timezone-unaware the `pandas_dtype` must be set directly on the Feature
            return "datetime64[ns, utc]"
        _log.info(
            f"Defaulting to pandas type 'object' for fqn {fqn} of type {typ.underlying.__name__}. Set the `pandas_dtype` attribute for better specificity."
        )
        return "object"

    def get_sqlalchemy_type(self, fqn: str) -> Union[TypeEngine, Type[TypeEngine]]:
        feature = self._fqn_to_feature[fqn]
        if feature.sqlalchemy_dtype is not None:
            return feature.sqlalchemy_dtype
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return Text
        if issubclass(underlying, bool):
            return Boolean
        if issubclass(underlying, int):
            return Integer
        if issubclass(underlying, float):
            return Float
        if issubclass(underlying, datetime):
            # This type requires the local timezone attribute to be set, but it will be stored as
            # a timezone-unaware timestamp in UTC. It will be converted back into a UTC tz-aware object
            # when parsed by python
            return UtcDateTime
        raise TypeError(
            f"Unable to infer a SQLAlchemy Dtype for field {fqn}. Please set the `sqlalchemy_dtype` attribute."
        )
